! Equations module for dark energy with constant equation of state parameter w
! allowing for perturbations based on a quintessence model
! by Antony Lewis (antony@camb.info)
! This version October 2002.


!I haven't really worked out a way of neatly getting round Fortrans module dependencies - 
!How do I define a lambda type here but also include it in ModelParams (which is used by
!the modules in this file?) For the moment, just set variable separately.

       module LambdaGeneral
         use precision
         real(dl)  :: w_lam = -1 !p/rho for the dark energy (assumed constant) 
         logical :: w_perturb = .true.


       end module LambdaGeneral


    
!Return OmegaK - modify this if you add extra fluid components
        function GetOmegak()
        use precision
        use ModelParams
        real(dl)  GetOmegak
         GetOmegak = 1 - (CP%omegab+CP%omegac+CP%omegav+CP%omegan) 
          
        end function GetOmegak
  
  
       subroutine init_background
         !This is only called once per model, and is a good point to do any extra initialization.
         !It is called before first call to dtauda, but after
         !massive neutrinos are initialized and after GetOmegak
       end  subroutine init_background


!Background evolution
        function dtauda(a)
         !get d tau / d a
        use precision
        use ModelParams
        use MassiveNu
        use LambdaGeneral
        implicit none
        real(dl) dtauda
        real(dl), intent(IN) :: a
        real(dl) rhonu,pnu,grhoa2, a2

        a2=a**2

!  8*pi*G*rho*a**4.
        grhoa2=grhok*a2+grhov*a**(1-3*w_lam)+(grhoc+grhob)*a+grhog+grhornomass
        if (CP%Num_Nu_massive /= 0) then
!Get massive neutrino density relative to massless
           call Nu_background(a*amnu,rhonu,pnu)
           grhoa2=grhoa2+rhonu*grhormass
        end if

        dtauda=sqrt(3/grhoa2)
     
        end function dtauda
!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

!Gauge-dependent perturbation equations

        module GaugeInterface
        use precision
        use ModelParams
        use MassiveNu
        use LambdaGeneral
        implicit none
        public

        !Description of this file. Change if you make modifications.
        character(LEN=*), parameter :: Eqns_name = 'gauge_inv'

        logical, parameter :: DoTensorNeutrinos = .false.

        integer, parameter :: max_l_evolve = 50 !Maximum l we are ever likely to propagate


        !Supported scalar initial condition flags
         integer, parameter :: initial_adiabatic=1, initial_iso_CDM=2, &
         initial_iso_baryon=3,  initial_iso_neutrino=4, initial_iso_neutrino_vel=5, initial_iso_quint=6,initial_vector = 0
         integer, parameter :: initial_nummodes = initial_iso_quint


        type EvolutionVars
        real(dl) q, q2
        real(dl) k_buf,k2_buf ! set in initial

        integer w_ix !Index of two quintessence equations

        integer q_ix !index into q_evolve array that gives the value q
        logical TransferOnly

!       nvar (t) - number of scalar (tensor) equations for this k       
        integer nvar,nvart

       !Max_l for the various hierarchies
        integer lmaxg,lmaxnr,lmaxnu,lmaxgpol,MaxlNeeded
        integer lmaxnrt, lmaxnut, lmaxt, lmaxpolt

        integer polind  !index into scalar array of polarization hierarchy

!array indices for massive neutrino equations
        integer iq0,iq1,iq2

!array index for tensor massive neutrino equations
        integer iqt
 
!Buffers for non-flat vars
        real(dl) Kf(max_l_evolve),Kft(max_l_evolve)      
   !Initial values for massive neutrino v*3 variables calculated when switching 
!to non-relativistic approx
        real(dl) G11,G30
        real(dl) w_nu !equation of state parameter for massive neutrinos
!True when using non-relativistic approximation
        logical MassiveNuApprox
  
!Numer of scalar equations we are propagating
        integer ScalEqsToPropagate
!beta > l for closed models 
        integer FirstZerolForBeta
!Tensor vars
        real(dl) tenspigdot, aux_buf

        real(dl) poltruncfac
        
  
        end type EvolutionVars
 

!precalculated arrays
        real(dl) polfac(max_l_evolve),tensfac(max_l_evolve),tensfacpol(max_l_evolve), &
               denl(max_l_evolve) 

        
       real(dl), parameter :: ep0=1.0d-2
    
       integer debug

        public GetNumEqns,output,outputt,initial,initialt,SwitchToMassiveNuApprox
        public outtransf
       contains


        subroutine GaugeInterface_Init
          !Precompute various arrays
          integer j
          
          if (CP%WantScalars) then
            do j=2,max_l_evolve
              polfac(j)=real((j+3)*(j-1),dl)/(j+1)
            end do               
          end if

          if (CP%WantTensors) then
           do j=2,max_l_evolve
           tensfac(j)=real((j+3)*(j-1),dl)/(j+1)
           tensfacpol(j)=tensfac(j)**2/(j+1)
          end do
          end if

         do j=1,max_l_evolve
           denl(j)=1._dl/(2*j+1)
         end do     
       
        end subroutine GaugeInterface_Init


        subroutine GetNumEqns(EV)
          use MassiveNu
          !Set the numer of equations in each hierarchy, and get total number of equations for this k
          type(EvolutionVars) EV
          real(dl) scal
        

         if (CP%Num_Nu_massive == 0) then
            EV%lmaxnu=0
         else 
            !l_max for massive neutrinos
             if (amnu < 2500 .and. EV%q > 0.02 ) then !Very light, so free-streaming more important             
               EV%lmaxnu=max(3,nint(6*lAccuracyBoost))   
             else
               EV%lmaxnu=max(3,nint(5*lAccuracyBoost))  
             end if
            if (CP%Transfer%high_precision) EV%lmaxnu=nint(25*lAccuracyBoost)
         end if

        if (CP%WantScalars) then
         EV%lmaxg  = max(nint(8*lAccuracyBoost),3)
         EV%lmaxnr = max(nint(7*lAccuracyBoost),3)
         EV%lmaxgpol = EV%lmaxg
         if (.not.CP%AccuratePolarization) EV%lmaxgpol=max(nint(4*lAccuracyBoost),3)
     
         if (EV%q < 0.05) then
            !Large scales need fewer equations
            scal  = 1
            if (CP%AccuratePolarization) scal = 4  !But need more to get polarization right
            EV%lmaxgpol=max(3,nint(min(8,nint(scal* 150* EV%q))*lAccuracyBoost))
            EV%lmaxnr=max(3,nint(min(7,nint(sqrt(scal)* 150 * EV%q))*lAccuracyBoost))
            EV%lmaxg=max(3,nint(min(8,nint(sqrt(scal) *300 * EV%q))*lAccuracyBoost))
         end if                  
         if (CP%Transfer%high_precision)  EV%lmaxnr=max(nint(25*lAccuracyBoost),3)
         EV%nvar=5+ (EV%lmaxg+1) + EV%lmaxgpol-1 +(EV%lmaxnr+1) 
         if (w_lam > -1 .and. w_Perturb) then
            EV%w_ix = EV%nvar+1
            EV%nvar=EV%nvar+2
         end if

         if (CP%Num_Nu_massive /= 0) then
           if (CP%MassiveNuMethod == Nu_approx) then
            EV%iq0=EV%nvar + 1 
            EV%nvar= EV%nvar+(EV%lmaxnu+1)
           else
            EV%iq0=EV%nvar + 1 
            EV%iq1=EV%iq0+nqmax
            EV%iq2=EV%iq1+nqmax
            EV%nvar=EV%nvar+ nqmax*(EV%lmaxnu+1)
           endif
         end if
        
         EV%MaxlNeeded=max(EV%lmaxg,EV%lmaxnr,EV%lmaxgpol,EV%lmaxnu)
         if (EV%MaxlNeeded > max_l_evolve) stop 'Need to increase max_l_evolve'
        
         EV%poltruncfac=real(EV%lmaxgpol,dl)/(EV%lmaxgpol-2)
         EV%polind = 6+EV%lmaxnr+EV%lmaxg
         EV%lmaxt=0
         
        else
          EV%nvar=0
        end if
    
       if (CP%WantTensors) then
          EV%lmaxt=max(3,nint(8*lAccuracyBoost))
          EV%lmaxpolt = max(3,nint(5*lAccuracyBoost)) 
          if (EV%q < 0.05) then
            !Large scales need fewer equations
            scal  = 1
            if (CP%AccuratePolarization) scal = 4 !But need more to get polarization right
            EV%lmaxt=max(3,nint(min(8,nint(scal *150 * EV%q))*lAccuracyBoost))
            EV%lmaxpolt=max(3,nint(min(5,nint( scal*150 * EV%q))*lAccuracyBoost))
          end if      
          EV%nvart=(EV%lmaxt-1)+(EV%lmaxpolt-1)*2+3
          if (DoTensorNeutrinos) then
           
            EV%lmaxnrt=nint(6*lAccuracyBoost)
            if (EV%q < 0.05) EV%lmaxnrt = max(3,nint(min(6,nint(150 * EV%q))*lAccuracyBoost)) 
            EV%lmaxnut=nint(4*lAccuracyBoost)
            EV%nvart=EV%nvart+EV%lmaxnrt-1
             if (CP%Num_Nu_massive /= 0 ) then
                 EV%iqt=EV%nvart + 1                  
                 EV%nvart=EV%nvart+ nqmax*(EV%lmaxnut-1)         
              end if 
          end if
        else
          EV%nvart=0
        end if       

        end subroutine GetNumEqns

!cccccccccccccccccccccccccccccccccc
        subroutine SwitchToMassiveNuApprox(EV,y)
!When the neutrinos are no longer highly relativistic we use a truncated
!energy-integrated hierarchy going up to third order in velocity dispersion
           type(EvolutionVars) EV

        real(dl) a,a2,pnu,clxnu,dpnu,pinu,grhormass_t,grhonu_t,rhonu
        real(dl) shearnu, qnu
        real(dl) y(EV%nvar)
   
        a=y(1)
        a2=a*a

        grhormass_t=grhormass/a2
        

           !Get density and pressure as ratio to massles by interpolation from table
           call Nu_background(a*amnu,rhonu,pnu)
           !Integrate over q

           call Nu_Integrate(a*amnu,clxnu,qnu,dpnu,shearnu,y(EV%iq0),y(EV%iq1),y(EV%iq2))
            !clxnu_here  = rhonu*clxnu, qnu_here = qnu*rhonu
            !Could save time by only calculating shearnu in output
           dpnu=dpnu/rhonu
           qnu=qnu/rhonu
           clxnu = clxnu/rhonu
           pinu=1.5_dl*shearnu/rhonu                     
    

          grhonu_t=grhormass_t*rhonu

          EV%MassiveNuApprox=.true.
        
          y(EV%iq0)=clxnu
          y(EV%iq0+1)=dpnu
          y(EV%iq0+2)=qnu
          y(EV%iq0+3)=pinu

          call Nu_Intvsq(a*amnu,EV%G11,EV%G30,y(EV%iq1),y(EV%iq0+3*nqmax))
!Analytic solution for higher moments, proportional to a^{-3}
          EV%G11=EV%G11*a**3/rhonu  
          EV%G30=EV%G30*a**3/rhonu  
        
          EV%ScalEqsToPropagate=EV%iq0+3

        end  subroutine SwitchToMassiveNuApprox


!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        subroutine output(EV,y, n,j,tau,sources)
        use ThermoData
        use lvalues
     
        implicit real(dl) (s-t)
        implicit logical (a-r,u-z)
        integer n,j
        type(EvolutionVars) EV
        real(dl), target :: y(n),yprime(n)
        real(dl), dimension(:),pointer :: ypol,ypolprime
        
        real(dl) dgq,grhob_t,grhor_t,grhoc_t,grhog_t,grhov_t,grhonu_t,gpnu_t,sigma,polter
        real(dl) qgdot,pigdot,pirdot,vbdot,dgrho
        real(dl) a,a2,z,clxc,clxb,vb,clxg,qg,pig,clxr,qr,pir,rhonu,pnu,pinu,clxnu,dpnu,qnu

        real(dl) tau,x,divfac,pinudot,rhonudot,shearnudot, grhormass_t
        real(dl) shearnu

        real(dl) k,k2  ,adotoa, grho, gpres,etak,phi
        real(dl) clv, clvdot, phidot
        real(dl) sources(NumScalSources)

        


        if (flat) then
            call fderivs(EV,EV%ScalEqsToPropagate,tau,y,yprime)
        else
            call derivs(EV,EV%ScalEqsToPropagate,tau,y,yprime)        
        end if
        
        ypolprime => yprime(EV%polind+1:)
        ypol => y(EV%polind+1:)

        k=EV%k_buf
        k2=EV%k2_buf
      
   
        a   =y(1)
        a2  =a*a
        etak=y(2)
        clxc=y(3)
        clxb=y(4)
        vb  =y(5)
        clxg=y(6)
        qg  =y(7)
        pig =y(8)
        clxr=y(7+EV%lmaxg)
        qr  =y(8+EV%lmaxg)
        pir =y(9+EV%lmaxg)
        pigdot=yprime(8)
        pirdot=yprime(EV%lmaxg+9)
        vbdot =yprime(5)
        qgdot =yprime(7)
    
!  Compute expansion rate from: grho 8*pi*rho*a**2

        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*a**(-1-3*w_lam)
        grhormass_t=grhormass/a2
        
        if (CP%Num_Nu_Massive == 0) then
           rhonu=1._dl
           pnu=1._dl/3._dl
           clxnu=0._dl
           pinu=0._dl
        else
           !Get density and pressure as ratio to massless by interpolation from table
           call Nu_background(a*amnu,rhonu,pnu)
   
           if (CP%MassiveNuMethod == Nu_approx) then

             clxnu=y(EV%iq0)
             qnu=y(EV%iq0+1)
             pinu=y(EV%iq0+2)
    
           else
           
           if (EV%MassiveNuApprox) then    
              clxnu=y(EV%iq0)
              !dpnu = y(EV%iq0+1)
              qnu=y(EV%iq0+2)
              pinu=y(EV%iq0+3)
           else
            !Integrate over q

            call Nu_Integrate(a*amnu,clxnu,qnu,dpnu,shearnu,y(EV%iq0),y(EV%iq1),y(EV%iq2))
            !clxnu_here  = rhonu*clxnu, qnu_here = qnu*rhonu
            !dpnu=dpnu/rhonu
            qnu=qnu/rhonu
            clxnu = clxnu/rhonu
            pinu=1.5_dl*shearnu/rhonu       
        endif
           end if
        
        end if

        EV%w_nu = pnu/rhonu


        grhonu_t=grhormass_t*rhonu
        gpnu_t=grhormass_t*pnu

        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t+grhonu_t
        gpres=(grhog_t+grhor_t)/3+grhov_t*w_lam +gpnu_t

        adotoa=sqrt((grho+grhok)/3)
   

!  8*pi*a*a*SUM[rho_i*clx_i]
        dgrho=grhob_t*clxb+grhoc_t*clxc + grhog_t*clxg+grhor_t*clxr +grhonu_t*clxnu

!  8*pi*a*a*SUM[(rho_i+p_i)*v_i]
        dgq=grhob_t*vb+grhog_t*qg+grhor_t*qr+ grhonu_t*qnu

        if (w_lam > -1 .and. w_Perturb) then
           clv=y(EV%w_ix)
           clvdot=y(EV%w_ix+1) 
           phidot = sqrt((1+w_lam)*grhov_t)
           dgrho=dgrho + (clvdot -clv*1.5_dl*(1-w_lam)*adotoa)*phidot
           dgq=dgq + k*phidot*clv
        end if

!  Get sigma (shear) and z from the constraints
!  have to get z from eta for numerical stability       
        z=(0.5_dl*dgrho/k + etak)/adotoa 
        sigma=(z+1.5_dl*dgq/k2)/EV%Kf(1)

 
        polter = 0.1_dl*pig+9._dl/15._dl*ypol(2)

        if (flat) then
        x=k*(tau0-tau)
        divfac=x*x    
        else   
        x=(tau0-tau)/r
        divfac=(r*rofChi(x))**2*k2 
        end if
       if (CP%Num_Nu_Massive == 0) then         
           pinudot=0._dl
        else
           if (CP%MassiveNuMethod == Nu_approx) then
              pinudot=yprime(EV%iq0+2)
           else
              if (EV%MassiveNuApprox) then
                 pinudot=yprime(EV%iq0+3)
              else
                 call Nu_derivs(a*amnu,adotoa,rhonu,rhonudot,shearnudot,y(EV%iq2),yprime(EV%iq2))
                 pinudot=1.5_dl*shearnudot/rhonu - rhonudot/rhonu*pinu    
              end if
           end if
        end if

!Maple's fortran output - see scal_eqs.map

      t1 = 1/adotoa
      t3 = adotoa**2
      t6 = 1/k
      t11 = grhonu_t*pinu
      t12 = grhor_t*pir
      t18 = k**2
      t19 = 1/t18
      t26 = t6/EV%Kf(1)
      s1 = (-t1*dgrho/3+(4*t3+grho/3+gpres)*sigma*t6+(-grhor_t*pirdot+(4 &
           *grhog_t*pig+3*pinu*gpnu_t+3*t11+4*t12)*adotoa-grhonu_t*pinudot- &
           grhog_t*pigdot)*t19+(2._dl/3._dl*sigma-2._dl/3._dl*t1*etak)*k-2*adotoa* &
           etak*t26)*expmmu(j)+((-9._dl/20._dl*ypolprime(2)-9._dl/160._dl*pigdot)* &
           t19*vis(j)+(-9._dl/160._dl*pig-9._dl/20._dl*ypol(2))*dvis(j)*t19)*opac(j)  
      t62 = s1+(pig/16+clxg/4+3._dl/8._dl*ypol(2)+21._dl/10._dl*t26*etak+(-21._dl &
           /5._dl*adotoa*sigma+(-9._dl/80._dl*yprime(9)-3._dl/8._dl*ypolprime(3) &
           )*EV%Kf(2)+3._dl/40._dl*qgdot+vbdot)*t6+(-21._dl/10._dl*t12-9._dl/20._dl* &
           dopac(j)*ypol(2)-21._dl/10._dl*t11+(-9._dl/160._dl*dopac(j)-21._dl/10._dl &
           *grhog_t)*pig)*t19)*vis(j)+(11._dl/10._dl*sigma+(-3._dl/8._dl*ypol(3) &
           -9._dl/80._dl*y(9))*EV%Kf(2)+3._dl/40._dl*qg+vb)*dvis(j)*t6+((9._dl/8._dl* &
           ypolprime(2)+3._dl/16._dl*pigdot)*dvis(j)+9._dl/8._dl*ddvis(j)*ypol(2)+ &
           3._dl/16._dl*ddvis(j)*pig)*t19

      sources(1)=t62

      if (x > 0._dl) then
         !E polarization source
           sources(2)=vis(j)*polter*(15._dl/8._dl)/divfac 
               !factor of four because no 1/16 later
        else
           sources(2)=0
        end if
 
      if (NumScalSources > 2) then
         !Get lensing sources
         !Can modify this here if you want to get power spectra for other tracer
       if (tau>taurend .and. tau0-tau > 0.1_dl) then
         
         !phi_lens = phi - 1/2 kappa (a/k)^2 sum_i rho_i pi_i
         phi = -(dgrho +3*dgq*adotoa/k)/(k2*EV%Kf(1)*2)  - (grhor_t*pirdot + grhog_t*pig+ pinu*gpnu_t)/k2
         
         sources(3) = -2*expmmu(j)*phi*(tau-tau_maxvis)/((tau0-tau_maxvis)*(tau0-tau))
       
          !We include the lensing factor of two here
       else
         sources(3) = 0
       end if
      end if

     end subroutine output


!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        subroutine outputt(EV,yt,n,j,tau,dt,dte,dtb)
!calculate the tensor sources for open and closed case
        use ThermoData

        implicit none
        integer j,n
        type(EvolutionVars) :: EV
        real(dl), target :: yt(n), ytprime(n)
        real(dl) tau,dt,dte,dtb,x,polterdot,polterddot,prefac
        real(dl) pig, pigdot, aux, polter, shear
        real(dl) sinhxr,cothxor
        real(dl) k,k2
        real(dl), dimension(:),pointer :: E,Bprime,Eprime

        if (flat) then
           call fderivst(EV,EV%nvart,tau,yt,ytprime)
        else
           call derivst(EV,EV%nvart,tau,yt,ytprime)
        end if
      
        k2=EV%k2_buf
        k=EV%k_buf 
        pigdot=EV%tenspigdot
        pig=yt(4)
        aux=EV%aux_buf  
        shear = yt(3)

        x=(tau0-tau)/r
       
        sinhxr=rofChi(x)*r
    
        if (EV%q*sinhxr > 1.e-8_dl) then  

        prefac=sqrt(EV%q2*r*r-Ksign)
        cothxor=cosfunc(x)/sinhxr
        E => yt(EV%lmaxt+3-1:)   
        Eprime=> ytprime(EV%lmaxt+3-1:) 
      
        Bprime => Eprime(EV%lmaxpolt:)

        polter = 0.1_dl*pig + 9._dl/15._dl*E(2)
        polterdot=9._dl/15._dl*Eprime(2) + 0.1_dl*pigdot
        polterddot = 9._dl/15._dl*(-dopac(j)*(E(2)-polter)-opac(j)*(  &
                   Eprime(2)-polterdot) + k*(2._dl/3._dl*Bprime(2)*aux - &
                   5._dl/27._dl*Eprime(3)*EV%Kft(2))) &
                   +0.1_dl*(k*(-ytprime(5)*EV%Kft(2)/3._dl + 8._dl/15._dl*ytprime(3)) - &
                   dopac(j)*(pig - polter) - opac(j)*(pigdot-polterdot))

        dt=(shear*expmmu(j) + (15._dl/8._dl)*polter*vis(j)/k)*r/sinhxr**2/prefac 
   
        dte=r*15._dl/8._dl/k/prefac* &
            ((ddvis(j)*polter + 2._dl*dvis(j)*polterdot + vis(j)*polterddot)  &
              + 4._dl*cothxor*(dvis(j)*polter + vis(j)*polterdot) - &
                   vis(j)*polter*(k2 -6*cothxor**2))
      
        dtb=15._dl/4._dl*EV%q*r/k/prefac*(vis(j)*(2._dl*cothxor*polter + polterdot) + dvis(j)*polter)
   
        else
        dt=0._dl
        dte=0._dl
        dtb=0._dl
        end if

        end subroutine outputt



!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        subroutine initial(EV,y, tau)
!  Initial conditions.
        implicit none
        type(EvolutionVars) EV
        real(dl) y(EV%nvar)
        real(dl) Rp15,tau,x,x2,x3,om,omtau, &
             Rc,Rb,Rv,Rg,grhonu,chi
        real(dl) k,k2 
        real(dl) a,a2, iqg
        integer l,i
        integer, parameter :: i_clxg=1,i_clxr=2,i_clxc=3, i_clxb=4, &
                i_qg=5,i_qr=6,i_vb=7,i_pir=8, i_eta=9, i_aj3r=10,i_clv=11,i_clvdot=12
        integer, parameter :: i_max = i_clvdot
        real(dl) initv(6,1:i_max), initvec(1:i_max)

           
        EV%ScalEqsToPropagate=EV%nvar
        if (flat) then
         EV%k_buf=EV%q
         EV%k2_buf=EV%q2     
         EV%Kf(1:EV%MaxlNeeded)=1._dl !initialize for flat case 
        else
         EV%k2_buf=EV%q2-curv
         EV%k_buf=sqrt(EV%k2_buf)
                    
         do l=1,EV%MaxlNeeded
           EV%Kf(l)=1._dl-curv*(l*(l+2))/EV%k2_buf           
         end do
        end if

        k=EV%k_buf
        k2=EV%k2_buf
   
        if (closed) then
           EV%FirstZerolForBeta = nint(EV%q*r) 
        else 
           EV%FirstZerolForBeta=l0max !a large number
        end if

   
!  k*tau, (k*tau)**2, (k*tau)**3
        x=k*tau
        x2=x*x
        x3=x2*x
        grhonu=grhormass+grhornomass
                                                        
        om = (grhob+grhoc)/sqrt(3*(grhog+grhonu))       
        omtau=om*tau
        Rv=grhonu/(grhonu+grhog)
        Rg = 1-Rv
        Rc=CP%omegac/(CP%omegac+CP%omegab)
        Rb=1-Rc
        Rp15=4*Rv+15


        if (CP%Scalar_initial_condition > initial_nummodes) &
          stop 'Invalid initial condition for scalar modes'

        a=tau*adotrad*(1+omtau/4)
        a2=a*a
        
        initv=0

!  Set adiabatic initial conditions

        chi=1  !Get transfer function for chi
        initv(1,i_clxg)=-chi*EV%Kf(1)/3*x2*(1-omtau/5)
        initv(1,i_clxr)= initv(1,i_clxg)
        initv(1,i_clxb)=0.75_dl*initv(1,i_clxg)
        initv(1,i_clxc)=initv(1,i_clxb)
        initv(1,i_qg)=initv(1,i_clxg)*x/9._dl
        initv(1,i_qr)=-chi*EV%Kf(1)*(4*Rv+23)/Rp15*x3/27
        initv(1,i_vb)=0.75_dl*initv(1,i_qg)
        initv(1,i_pir)=chi*4._dl/3*x2/Rp15*(1+omtau/4*(4*Rv-5)/(2*Rv+15))
        initv(1,i_aj3r)=chi*4/21._dl/Rp15*x3
        initv(1,i_eta)=-chi*2*EV%Kf(1)*(1 - x2/12*(-10._dl/Rp15 + EV%Kf(1)))

        if (CP%Scalar_initial_condition/= initial_adiabatic) then
!CDM isocurvature   
       
         initv(2,i_clxg)= Rc*omtau*(-2._dl/3 + omtau/4)
         initv(2,i_clxr)=initv(2,i_clxg)
         initv(2,i_clxb)=initv(2,i_clxg)*0.75_dl
         initv(2,i_clxc)=1+initv(2,i_clxb)
         initv(2,i_qg)=-Rc/9*omtau*x
         initv(2,i_qr)=initv(2,i_qg)
         initv(2,i_vb)=0.75_dl*initv(2,i_qg)
         initv(2,i_pir)=-Rc*omtau*x2/3/(2*Rv+15._dl)
         initv(2,i_eta)= Rc*omtau*(1._dl/3 - omtau/8)*EV%Kf(1)
         initv(2,i_aj3r)=0
!Baryon isocurvature
         if (Rc==0) stop 'Isocurvature initial conditions assume non-zero dark matter'

         initv(3,:) = initv(2,:)*(Rb/Rc)
         initv(3,i_clxc) = initv(3,i_clxb)
         initv(3,i_clxb)= initv(3,i_clxb)+1
      
!neutrino isocurvature density mode
       
         initv(4,i_clxg)=Rv/Rg*(-1 + x2/6)
         initv(4,i_clxr)=1-x2/6
         initv(4,i_clxc)=-omtau*x2/80*Rv*Rb/Rg
         initv(4,i_clxb)= Rv/Rg/8*x2
         iqg = - Rv/Rg*(x/3 - Rb/4/Rg*omtau*x)
         initv(4,i_qg) =iqg
         initv(4,i_qr) = x/3
         initv(4,i_vb)=0.75_dl*iqg
         initv(4,i_pir)=x2/Rp15
         initv(4,i_eta)=EV%Kf(1)*Rv/Rp15/3*x2
         initv(6,i_aj3r)=0

!neutrino isocurvature velocity mode

         initv(5,i_clxg)=Rv/Rg*x - 2*x*omtau/16*Rb*(2+Rg)/Rg**2
         initv(5,i_clxr)=-x -3*x*omtau*Rb/16/Rg
         initv(5,i_clxc)=-9*omtau*x/64*Rv*Rb/Rg
         initv(5,i_clxb)= 3*Rv/4/Rg*x - 9*omtau*x/64*Rb*(2+Rg)/Rg**2
         iqg = Rv/Rg*(-1 + 3*Rb/4/Rg*omtau+x2/6 +3*omtau**2/16*Rb/Rg**2*(Rg-3*Rb))
         initv(5,i_qg) =iqg
         initv(5,i_qr) = 1 - x2/6*(1+4*EV%Kf(1)/(4*Rv+5))
         initv(5,i_vb)=0.75_dl*iqg
         initv(5,i_pir)=2*x/(4*Rv+5)+omtau*x*6/Rp15/(4*Rv+5)
         initv(5,i_eta)=2*EV%Kf(1)*x*Rv/(4*Rv+5) + omtau*x*3*EV%Kf(1)*Rv/32*(Rb/Rg - 80/Rp15/(4*Rv+5))
         initv(5,i_aj3r) = 3._dl/7*x2/(4*Rv+5)

!quintessence isocurvature mode

         if (w_lam > -1 .and. w_Perturb) then
            !This doesn't make much sense as clv is dimensional,
            !and w_lam /= -1 is not consistent with the series assumptions...
            initv(6,i_clv) = 1 - x2/6+omtau*x2/72 
            initv(6,i_clvdot) = k*(-x/3 + omtau*x/24)
   
         end if 

         end if

   if (CP%Scalar_initial_condition==initial_vector) then
          InitVec = 0
          do i=1,initial_nummodes
          InitVec = InitVec+ initv(i,:)*CP%InitialConditionVector(i)
          end do
       else
          InitVec = initv(CP%Scalar_initial_condition,:)
          if (CP%Scalar_initial_condition==initial_adiabatic) InitVec = -InitVec
            !So we start with chi=-1 as before
       end if

        y(1)=a
        y(2)= -InitVec(i_eta)*k/2
        !get eta_s*k, where eta_s is synchronous gauge variable

!  CDM
        y(3)=InitVec(i_clxc)

!  Baryons
        y(4)=InitVec(i_clxb)
        y(5)=InitVec(i_vb)

!  Photons
        y(6)=InitVec(i_clxg)
        y(7)=InitVec(i_qg) 
      
        
        y(8:EV%nvar)=0._dl
 
        if (w_lam > -1 .and. w_Perturb) then
         y(EV%w_ix) = InitVec(i_clv)
         y(EV%w_ix+1) = InitVec(i_clvdot)
        end if

!  Neutrinos
        y(7+EV%lmaxg)=InitVec(i_clxr)
        y(8+EV%lmaxg)=InitVec(i_qr)
        y(9+EV%lmaxg)=InitVec(i_pir)

        if (EV%FirstZerolForBeta==3) then
         y(10+EV%lmaxg)=0._dl
        else
         y(10+EV%lmaxg)=InitVec(i_aj3r)
        endif

        EV%MassiveNuApprox=.false.
        if (CP%Num_Nu_massive == 0) return 
       
        if (CP%MassiveNuMethod == Nu_approx) then
         !Values are just same as massless neutrino ones since highly relativistic
         y(EV%iq0)=InitVec(i_clxr)
         y(EV%iq0+1)=InitVec(i_qr)
         y(EV%iq0+2)=InitVec(i_pir)
          if (EV%FirstZerolForBeta/=3) then
           y(EV%iq0+3)=InitVec(i_aj3r)
          endif
        else
         do  i=1,nqmax
          y(EV%iq0+i-1)=-0.25_dl*dlfdlq(i)*InitVec(i_clxr)
          y(EV%iq1+i-1)=-0.25_dl*dlfdlq(i)*InitVec(i_qr)
          y(EV%iq2+i-1)=-0.25_dl*dlfdlq(i)*InitVec(i_pir)
         end do       
        end if  
       
        end subroutine initial


!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        subroutine initialt(EV,yt,tau)
!  Initial conditions for tensors

        implicit none
        real(dl) bigR,tau,x,aj3r,elec, pir
        integer l
         type(EvolutionVars) EV
        real(dl) k,k2 ,a
        real(dl) yt(EV%nvart)
        
        if (flat) then
         EV%aux_buf=1._dl
         EV%k2_buf=EV%q2
         EV%k_buf=EV%q       
         EV%Kft(1:EV%lmaxt)=1._dl!initialize for flat case
        else
      
        EV%k2_buf=EV%q2-3*curv
        EV%k_buf=sqrt(EV%k2_buf) 
        EV%aux_buf=sqrt(1._dl+3*curv/EV%k2_buf)  
        do l=1,EV%lmaxt
           EV%Kft(l)=1._dl-curv*((l+1)**2-3)/EV%k2_buf
        end do
        endif
        
        k=EV%k_buf
        k2=EV%k2_buf

        if (closed) then
           EV%FirstZerolForBeta = nint(EV%q*r) 
        else 
           EV%FirstZerolForBeta=l0max !a large number
        end if
         
        a=tau*adotrad
        
        if (DoTensorNeutrinos) then
         bigR = (grhormass+grhornomass)/(grhormass+grhornomass+grhog)
        else
         bigR = 0._dl
        end if

        yt(1)=a
        
        yt(2)= 1._dl
        
        elec=-(1+2*curv/k2)*(2*bigR+10)/(4*bigR+15) !elec, with H=1
    
        x=k*tau
       
        !shear
        yt(3)=-5._dl/2/(bigR+5)*x*elec 
        
        yt(4:EV%nvart)=0._dl
        
!  Neutrinos 
        if (DoTensorNeutrinos) then
         pir=-2._dl/3._dl/(bigR+5)*x**2*elec
         aj3r=  -2._dl/21._dl/(bigR+5)*x**3*elec
         yt((EV%lmaxt-1)+(EV%lmaxt-1)*2+3+1)=pir
         yt((EV%lmaxt-1)+(EV%lmaxt-1)*2+3+2)=aj3r
        end if
     
    
        end subroutine initialt


      subroutine outtransf(EV, y, Arr)
 !write out clxc, clxb, clxg, clxn
        use Transfer
        implicit none
        type(EvolutionVars) EV
   
        real(dl) clxc, clxb, clxg, clxr, clxnu, clxtot,k,k2
        real Arr(Transfer_max)
        real(dl) y(EV%nvar)
        real(dl) qnu, a, rhonu, pnu

        a    = y(1)
        clxc = y(3)
        clxb = y(4)
        clxg = y(6)
        clxr = y(7+EV%lmaxg)
        k    = EV%k_buf
        k2   = EV%k2_buf
         
        if (CP%Num_Nu_Massive > 0) then
           call Nu_background(a*amnu,rhonu,pnu)

            if (EV%MassiveNuApprox) then    
              clxnu=y(EV%iq0)
            else
            !Integrate over q
            call Nu_Integrate01(a*amnu,clxnu,qnu,y(EV%iq0),y(EV%iq1))
            clxnu=clxnu/rhonu
           endif

        else
           clxnu = 0
           rhonu  = 0
        end if

        clxtot = (clxc*grhoc/a + clxb*grhob/a + clxnu*grhormass/a**2*rhonu)
        clxtot=clxtot/(grhoc/a+grhob/a + grhormass/a**2*rhonu)
        
        Arr(Transfer_kh) = k/(CP%h0/100._dl)
        Arr(Transfer_cdm) = clxc/k2
        Arr(Transfer_b) = clxb/k2
        Arr(Transfer_g) = clxg/k2
        Arr(Transfer_r) = clxr/k2
        Arr(Transfer_nu) = clxnu/k2
        Arr(Transfer_tot) = clxtot/k2

     end subroutine outtransf

!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        subroutine fderivs(EV,n,tau,ay,ayprime)
!  Evaluate the time derivatives of the perturbations, flat case
!  ayprime is not neccessarily GaugeInterface.yprime, so keep them distinct
        use ThermoData
        use MassiveNu
        implicit none      
        type(EvolutionVars) EV

        integer n
        real(dl) ay(n),ayprime(n)
        real(dl) tau,ep,w
        real(dl) k,k2 

!  Internal variables.

        real(dl) opacity

        real(dl) photbar,cs2,pb43,grho,slip,clxgdot,grhormass_t, &
                      clxcdot,clxbdot,drag,adotdota,gpres,clxrdot,etak
        real(dl) q,aq,v,akv(nqmax0)
        real(dl) G11_t,G30_t

        real(dl) dgq,grhob_t,grhor_t,grhoc_t,grhog_t,grhov_t,grhonu_t,gpnu_t,sigma,polter
        real(dl) qgdot,qrdot,pigdot,pirdot,vbdot,dgrho,adotoa
        real(dl) a,a2,z,clxc,clxb,vb,clxg,qg,pig,clxr,qr,pir
        real(dl) rhonu,pnu,clxnu,qnu,f
        real(dl) phidot, clv, clvdot, Vphiphi

        integer l,i,ind
        logical tcp

                    
        k=EV%k_buf
        k2=EV%k2_buf     
       

        a=ay(1)
        etak=ay(2)


!  CDM variables
        clxc=ay(3)

!  Baryon variables
        clxb=ay(4)
        vb=ay(5)

!  Photons
        clxg=ay(6)
        qg=ay(7)
        pig=ay(8)

!  Massless neutrinos
        clxr=ay(7+EV%lmaxg)
        qr=ay(8+EV%lmaxg)
        pir=ay(9+EV%lmaxg)

        a2=a*a

!  Get sound speed and ionisation fraction.
        call thermo(tau,cs2,opacity)

!  Compute expansion rate from: grho 8*pi*rho*a**2

        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*a**(-1-3*w_lam)
        grhormass_t=grhormass/a2
       

        if (CP%Num_Nu_Massive == 0) then
           rhonu=1
           pnu=1._dl/3._dl
           clxnu=0
           qnu = 0
        else 
               
           !Get density and pressure as ratio to massles by interpolation from table
           call Nu_background(a*amnu,rhonu,pnu)
        
           if (CP%MassiveNuMethod == Nu_approx) then
             clxnu=ay(EV%iq0)
             qnu=ay(EV%iq0+1)
           else
           if (EV%MassiveNuApprox) then    
              clxnu=ay(EV%iq0)
              qnu=ay(EV%iq0+2)
            else
            !Integrate over q

            call Nu_Integrate01(a*amnu,clxnu,qnu,ay(EV%iq0),ay(EV%iq1))
            !clxnu_here  = rhonu*clxnu, qnu_here = qnu*rhonu
            qnu=qnu/rhonu
            clxnu = clxnu/rhonu
           endif 
           end if

        end if


        grhonu_t=grhormass_t*rhonu
        gpnu_t=grhormass_t*pnu

        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t+grhonu_t
        
        adotoa=sqrt(grho/3)
        ayprime(1)=adotoa*a

!  Photon mass density over baryon mass density
        photbar=grhog_t/grhob_t
        pb43=4._dl/3._dl*photbar

!  Evaluate the time derivative of sz (gradient of expansion)
!  8*pi*a*a*SUM[rho_i*clx_i]
        dgrho=grhob_t*clxb+grhoc_t*clxc + grhog_t*clxg+grhor_t*clxr +grhonu_t*clxnu
       
!  8*pi*a*a*SUM[(rho_i+p_i)*v_i]
        dgq=grhob_t*vb+grhog_t*qg+grhor_t*qr + grhonu_t*qnu
       
        if (w_lam > -1 .and. w_Perturb) then
           clv=ay(EV%w_ix)
           clvdot=ay(EV%w_ix+1) 
           phidot = sqrt((1+w_lam)*grhov_t)
           dgrho=dgrho + (clvdot -clv*1.5_dl*(1-w_lam)*adotoa)*phidot
           dgq=dgq + k*phidot*clv
        end if
!  Get sigma (shear) and z from the constraints
! have to get z from eta for numerical stability
        z=(0.5_dl*dgrho/k + etak)/adotoa 
        sigma=z+1.5_dl*dgq/k2
        !  ddota/a
        gpres=(grhog_t+grhor_t)/3._dl +gpnu_t +grhov_t*w_lam
        adotdota=0.5_dl*(adotoa*adotoa-gpres)

        if (w_lam > -1 .and. w_Perturb) then
           ayprime(EV%w_ix)= clvdot
! a^2 d^2 V/d phi^2
           Vphiphi = -1.5_dl*(1-w_lam)*(adotdota - (7+3*w_lam)/2*adotoa**2)
           ayprime(EV%w_ix+1) = - 2*adotoa*clvdot - k*z*phidot - k2*clv - clv*Vphiphi
        end if

!eta*k equation
        ayprime(2)=0.5_dl*dgq

!  CDM equation of motion
        clxcdot=-k*z
        ayprime(3)=clxcdot

!  Baryon equation of motion.
        clxbdot=-k*(z+vb)
        ayprime(4)=clxbdot
!  Photon equation of motion
        clxgdot=-k*(4._dl/3._dl*z+qg)

!  Drag: a*n_e*sigma_T*(4/3*vb-qg)
        drag=opacity*(4._dl/3._dl*vb-qg)
       
!  use tight coupling approx?
        if (k > epsw) then
           ep=ep0
        else
           ep=0.5_dl*ep0
        end if
!     There was an instability for low k when we 
!     switched off tight coupling. Minor change in 
!     criteria ends the problem. (Thanks to David Spergel)      
        tcp = ((k/opacity > ep).or.(1._dl/(opacity*tau) > ep .and. k/opacity > 1d-4)) 
       
!  Use explicit equation for vb if appropriate

         if (tcp) then
           vbdot=-adotoa*vb+cs2*k*clxb-photbar*drag
        else

!  Use tight-coupling approximation for vb
!  zeroth order (in t_c) approximation to vbdot
        vbdot=(-adotoa*vb+cs2*k*clxb  &
             +0.25_dl*k*pb43*(clxg-2._dl*pig))/(1._dl+pb43)



!  First-order (in t_c) approximation to baryon-photon splip
        slip=2._dl*pb43/(1._dl+pb43)*adotoa*(vb-3._dl/4._dl*qg) &
           +1._dl/opacity*(-adotdota*vb-0.5_dl*k*adotoa*clxg  &
           +k*(cs2*clxbdot-0.25_dl*clxgdot))/(1._dl+pb43)

!  First-order approximation to vbdot
        vbdot=vbdot+pb43/(1._dl+pb43)*slip
        end if
        ayprime(5)=vbdot

!  Photon equations of motion
        ayprime(6)=clxgdot
        qgdot=4._dl/3._dl*(-vbdot-adotoa*vb+cs2*k*clxb)/pb43 &
            +k*(clxg-2._dl*pig)/3._dl
        ayprime(7)=qgdot
        
        polter = 0.1_dl*pig+9._dl/15._dl*ay(EV%polind+2) !2/15*(3/4 pig + 9/2 E2)
!  Use explicit equations for photon moments if appropriate       
        if (tcp) then
        pigdot=0.4_dl*k*qg-0.6_dl*k*ay(9)-opacity*(pig - polter) &
               +8._dl/15._dl*k*sigma
        ayprime(8)=pigdot
        do  l=3,EV%lmaxg-1
           ayprime(l+6)=k*denl(l)*(l*ay(l+5)-(l+1)*ay(l+7))-opacity*ay(l+6)
        end do
!  Truncate the photon moment expansion
        ayprime(EV%lmaxg+6)=k*ay(EV%lmaxg+5)-(EV%lmaxg+1)/tau*ay(EV%lmaxg+6)  &
                       -opacity*ay(EV%lmaxg+6)
        else
!  Use tight-coupling approximation where moments are zero for l>1
        pigdot=0._dl
        
        ayprime(8:EV%lmaxg+6)=0._dl
        
        end if

!  Massless neutrino equations of motion.
        clxrdot=-k*(4._dl/3._dl*z+qr)
        ayprime(EV%lmaxg+7)=clxrdot
        qrdot=k*(clxr-2._dl*pir)/3._dl
        ayprime(EV%lmaxg+8)=qrdot
        pirdot=k*(0.4_dl*qr-0.6_dl*ay(EV%lmaxg+10)+8._dl/15._dl*sigma)
        ayprime(EV%lmaxg+9)=pirdot
        do l=3,EV%lmaxnr-1
           ayprime(l+EV%lmaxg+7)=k*denl(l)*(l*ay(l+EV%lmaxg+6) -(l+1)*ay(l+EV%lmaxg+8))
        end do   
!  Truncate the neutrino expansion
        ayprime(EV%lmaxnr+EV%lmaxg+7)=k*ay(EV%lmaxnr+EV%lmaxg+6)- &
                                (EV%lmaxnr+1)/tau*ay(EV%lmaxnr+EV%lmaxg+7)
     
!  Polarization
        if (tcp) then     
        !l=2 (defined to be zero for l<2)
        ayprime(EV%polind+2) = -opacity*(ay(EV%polind+2) - polter) - k/3._dl*ay(EV%polind+3)
        !and the rest
        do l=3,EV%lmaxgpol-1
           ayprime(EV%polind+l)=-opacity*ay(EV%polind+l) + k*denl(l)*(l*ay(EV%polind+l-1) -&
                          polfac(l)*ay(EV%polind+l+1))
        end do  
        
        !truncate
        ayprime(EV%polind+EV%lmaxgpol)=-opacity*ay(EV%polind+EV%lmaxgpol) + &
           k*EV%poltruncfac*ay(EV%polind+EV%lmaxgpol-1)-(EV%lmaxgpol+3)*ay(EV%polind+EV%lmaxgpol)/tau

        else
         ayprime(EV%polind+2:EV%polind+EV%lmaxgpol)=0._dl
        end if

!  Massive neutrino equations of motion.
         if (CP%Num_Nu_massive == 0) return

         if (CP%MassiveNuMethod == Nu_approx) then

            !crude approximation scheme
            w=pnu/rhonu
           
            f = (5._dl/3)**(a*amnu/(a*amnu+200))*(3*w)**((2+a*amnu)/(4+a*amnu)) 

            !clxnudot
            ayprime(EV%iq0)=-(f-3*w)*adotoa*clxnu-k*((1+w)*z+qnu)
            !qnudot
            ayprime(EV%iq0+1)=-adotoa*(1-3*w)*ay(EV%iq0+1)+k/3._dl*(f*clxnu-2._dl*ay(EV%iq0+2))
            !pinudot

            ayprime(EV%iq0+2)=-adotoa*(-f +2-3*w)*ay(EV%iq0+2) +  &
                 k*(2._dl/5*f*qnu-0.6_dl*ay(EV%iq0+3)+ 2._dl/5*w*(5-3*w)*sigma)
 
            do l=3,EV%lmaxnu-1
               ayprime(EV%iq0+l)=-adotoa*((1-l)*f+l-3*w)*ay(EV%iq0+l) + &
                    k*denl(l)*(f*l*ay(EV%iq0+l-1) -(l+1)*ay(EV%iq0+l+1))
            end do
            !  Truncate the neutrino expansion
            ayprime(EV%iq0+EV%lmaxnu)=k*ay(EV%iq0+EV%lmaxnu-1)- &
                 (EV%lmaxnu+1)/tau*ay(EV%iq0+EV%lmaxnu)
                        
                     
         else !Not approx scheme

          if (EV%MassiveNuApprox) then
             !Now EV%iq0 = clx, EV%iq0+1 = clxp, EV%iq0+2 = G_1, EV%iq0+3=G_2=pinu
             G11_t=EV%G11/a/a2 
             G30_t=EV%G30/a/a2
             w=pnu/rhonu
             ayprime(EV%iq0)=-k*z*(w+1) + 3*adotoa*(w*ay(EV%iq0) - ay(EV%iq0+1))-k*ay(EV%iq0+2)
             ayprime(EV%iq0+1)=(3*w-2)*adotoa*ay(EV%iq0+1) - 5._dl/3*k*z*w - k/3*G11_t
             ayprime(EV%iq0+2)=(3*w-1)*adotoa*ay(EV%iq0+2) - k*(2._dl/3*ay(EV%iq0+3)-ay(EV%iq0+1))
             ayprime(EV%iq0+3)=(3*w-2)*adotoa*ay(EV%iq0+3) + 2*w*k*sigma - k/5*(3*G30_t-2*G11_t)
     
          else
          
          do i=1,nqmax
            q=i*dq-0.5_dl
            aq=a*amnu/q
            v=1._dl/sqrt(1._dl+aq*aq)
            akv(i)=k*v        
          end do
        
!  l = 0, 1, 2,EV%lmaxnu.
          do i=1,nqmax
           ind=EV%iq0+i-1
           ayprime(ind)=-akv(i)*ay(ind+nqmax)+z*k*dlfdlq(i)/3
           ind=EV%iq1+i-1
           ayprime(ind)=akv(i)*(ay(ind-nqmax)-2*ay(ind+nqmax))/3
           ind=EV%iq2+i-1
           ayprime(ind)=akv(i)*(2*ay(ind-nqmax)-3*ay(ind+nqmax))/5 &
                 -k*2._dl/15._dl*sigma*dlfdlq(i)
           ind=EV%iq0+i-1+EV%lmaxnu*nqmax
!  Truncate moment expansion.
           ayprime(ind)=akv(i)*ay(ind-nqmax)-(EV%lmaxnu+1)/tau*ay(ind)
          end do
          do l=3,EV%lmaxnu-1
            do i=1,nqmax
            ind=EV%iq0-1+i+l*nqmax
            ayprime(ind)=akv(i)*denl(l)*(l*ay(ind-nqmax)-(l+1)*ay(ind+nqmax))
            end do
          end do

          end if

          end if

        end subroutine fderivs

!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        subroutine fderivst(EV,n,tau,ayt,aytprime)
!  Evaluate the time derivatives of the tensor perturbations, flat case
        use ThermoData
        use MassiveNu
        implicit none      
        type(EvolutionVars) EV
        integer n,l,i,ind
        real(dl), target ::  ayt(n),aytprime(n)
        real(dl) ep,tau,grho,rhopi,cs2,opacity,gpres,pirdt,grhormass_t
        logical tcp
        real(dl), dimension(:),pointer :: neut,neutprime,E,B,Eprime,Bprime
        real(dl) q,aq,v,akv(nqmax0),shearnu
        real(dl)  grhob_t,grhor_t,grhoc_t,grhog_t,grhov_t,grhonu_t,gpnu_t,polter
        real(dl) Hchi,shear, pig
        real(dl) k,k2,a,a2
        real(dl) pinu,pir,pnu,adotoa, rhonu

         k2=EV%k2_buf
         k=EV%k_buf       

        !E and B start at l=2. Set up pointers accordingly to fill in y arrays
        E => ayt(EV%lmaxt+3-1:)
        Eprime=> aytprime(EV%lmaxt+3-1:) 
        B => E(EV%lmaxpolt:)
        Bprime => Eprime(EV%lmaxpolt:)
       
     
        
        a=ayt(1)        

!Hchi = metric perturbation variable, conserved on large scales
        Hchi=ayt(2)

!shear (Hchi' = - k*shear)
        shear=ayt(3)

!  Photon anisotropic stress
        pig=ayt(4)

        
        a2=a*a

!  Get sound speed and opacity, and see if should use tight-coupling
             
        call thermo(tau,cs2,opacity)
        if (k > 0.06_dl*epsw) then
           ep=ep0
        else
           ep=0.2_dl*ep0
        end if

        tcp = ((k/opacity > ep).or.(1._dl/(opacity*tau) > ep)) 

!Do massive neutrinos
        if (CP%Num_Nu_Massive ==0) then
           rhonu=1._dl
           pnu=1._dl/3._dl
        else
           call Nu_background(a*amnu,rhonu,pnu)

        end if

! Compute expansion rate from: grho=8*pi*rho*a**2
! Also calculate gpres: 8*pi*p*a**2
        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*a**(-1-3*w_lam)
        grhormass_t=grhormass/a2

        grhonu_t=grhormass_t*rhonu
        gpnu_t=grhormass_t*pnu

        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t+grhonu_t
       
        gpres=(grhog_t+grhor_t)/3._dl+grhov_t*w_lam +gpnu_t
        adotoa=sqrt(grho/3._dl)
       
        aytprime(1)=adotoa*a
        polter = 0.1_dl*pig + 9._dl/15._dl*E(2)

     

        if (tcp) then
!  Use explicit equations:
!  Equation for the photon anisotropic stress
        aytprime(4)=k*(-1._dl/3._dl*ayt(5)+8._dl/15._dl*shear)  &
                  -opacity*(pig - polter)
! And for the moments            
        do  l=3,EV%lmaxt-1
           aytprime(l+2)=k*denl(l)*(l*ayt(l+1)-   &
                  tensfac(l)*ayt(l+3))-opacity*ayt(l+2)
        end do
!  Truncate the hierarchy
        
        aytprime(EV%lmaxt+2)=k*EV%lmaxt/(EV%lmaxt-2._dl)*ayt(EV%lmaxt+1)- &
                       (EV%lmaxt+3._dl)*ayt(EV%lmaxt+2)/tau-opacity*ayt(EV%lmaxt+2)
     
!E equations

        Eprime(2) = - opacity*(E(2) - polter) + k*(4._dl/6._dl*B(2) - &
                        5._dl/27._dl*E(3))
        do l=3,EV%lmaxpolt-1
        Eprime(l) =-opacity*E(l) + k*(denl(l)*(l*E(l-1) - &
                        tensfacpol(l)*E(l+1)) + 4._dl/(l*(l+1))*B(l))
        end do
!truncate
        Eprime(EV%lmaxpolt)=0._dl
        
        
!B-bar equations
        
        do l=2,EV%lmaxpolt-1
        Bprime(l) =-opacity*B(l) + k*(denl(l)*(l*B(l-1) - &
                        tensfacpol(l)*B(l+1)) - 4._dl/(l*(l+1))*E(l))
        end do
!truncate
        Bprime(EV%lmaxpolt)=0._dl

       else

        ayt(4)=32._dl/45._dl*k/opacity*shear   
        E(2)=ayt(4)/4._dl
!  Set the derivatives to zero
        aytprime(4:n)=0._dl
        
        endif
     

!  Neutrino equations: 
!  Anisotropic stress  
        if (DoTensorNeutrinos) then
        
        neutprime => Bprime(EV%lmaxpolt:)
        neut => B(EV%lmaxpolt:)
               
!  Massless neutrino anisotropic stress
        pir=neut(2)
      
        pirdt=-1._dl/3._dl*k*neut(3)+ 8._dl/15._dl*k*shear
        neutprime(2)=pirdt
!  And for the moments
        do  l=3,EV%lmaxnrt-1
           neutprime(l)=k*denl(l)*(l*neut(l-1)- tensfac(l)*neut(l+1))
        end do
        
!  Truncate the hierarchy
        neutprime(EV%lmaxnrt)=k*EV%lmaxnrt/(EV%lmaxnrt-2._dl)*neut(EV%lmaxnrt-1)-  &
                       (EV%lmaxnrt+3._dl)*neut(EV%lmaxnrt)/tau
    
        !  Massive neutrino equations of motion.
         if (CP%Num_Nu_massive /= 0) then
          
          do i=1,nqmax
            q=i*dq-0.5_dl
            aq=a*amnu/q
            v=1._dl/sqrt(1._dl+aq*aq)
            akv(i)=k*v
          end do
          do i=1,nqmax
           ind=EV%iqt+i-1                          
           aytprime(ind)=-1._dl/3._dl*akv(i)*ayt(ind+nqmax)- 2._dl/15._dl*k*shear*dlfdlq(i)
           ind=EV%iqt+i-1+(EV%lmaxnut-2)*nqmax
!  Truncate moment expansion.
           aytprime(ind)=akv(i)*EV%lmaxnut/(EV%lmaxnut-2._dl)*ayt(ind-nqmax)-(EV%lmaxnut+3)/tau*ayt(ind)
          end do
          do l=3,EV%lmaxnut-1
            do i=1,nqmax
            ind=EV%iqt-1+i+(l-2)*nqmax
            aytprime(ind)=akv(i)*denl(l)*(l*ayt(ind-nqmax)-tensfac(l)*ayt(ind+nqmax))     
           end do
          end do

           call Nu_Shear(a*amnu,shearnu,ayt(EV%iqt))                 
           pinu=1.5_dl*shearnu/rhonu   
  
         else
            pinu=0._dl
         end if
              
        else
          pinu=0
          pir=0
          pirdt=0
        end if
      
!  Get the propagation equation for the shear
        
            
        rhopi=grhog_t*pig+grhor_t*pir+grhonu_t*pinu
                         
        aytprime(3)=-2*adotoa*shear+k*Hchi-rhopi/k   

        if (tcp) then
!  Use the full expression for pigdt
           EV%tenspigdot=aytprime(4)
        else
!  Use the tight-coupling approximation
           EV%tenspigdot= 32._dl/45._dl*k/opacity*(2._dl*adotoa*shear+aytprime(3))
         endif

        aytprime(2)=-k*shear


end subroutine fderivst
      

!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        subroutine derivs(EV,n,tau,ay,ayprime)
!  Evaluate the time derivatives of the perturbations.
!ayprime is not neccessarily GaugeInterface.yprime, so keep them distinct
        use ThermoData
        use MassiveNu
        implicit none      
        type(EvolutionVars) EV
        integer n
        real(dl) ay(n),ayprime(n)
        real(dl) tau

        real(dl) dgq,grhob_t,grhor_t,grhoc_t,grhog_t,grhov_t,grhonu_t,gpnu_t,sigma,polter
        real(dl) qgdot,qrdot,pigdot,pirdot,vbdot,dgrho,adotoa
        real(dl) a,a2,z,clxc,clxb,vb,clxg,qg,pig,clxr,qr,pir,rhonu,pnu,clxnu,qnu
        real(dl) k,k2 
!  Internal variables.

        real(dl) opacity,ep,cothxor,w

        real(dl) photbar,cs2,pb43,grho,slip,clxgdot,etak, &
                        clxcdot,clxbdot,drag,adotdota,gpres,clxrdot
        real(dl) q,aq,v,akv(nqmax0),grhormass_t
        integer l,i,ind
        logical tcp !false if can use tight coupling approx
   
        real(dl) G11_t,G30_t, f
        real(dl) clv,clvdot, phidot, Vphiphi


   
        k2=EV%k2_buf
        k=EV%k_buf


        cothxor=1._dl/tanfunc(tau/r)/r      
      
        a=ay(1)
        !z=ay(2)/a
        etak=ay(2)

!  CDM variables
        clxc=ay(3)

!  Baryon variables
        clxb=ay(4)
        vb=ay(5)

!  Photons
        clxg=ay(6)
        qg=ay(7)
        pig=ay(8)

!  Massless neutrinos
        clxr=ay(7+EV%lmaxg)
        qr=ay(8+EV%lmaxg)
        pir=ay(9+EV%lmaxg)

        a2=a*a

!  Get baryon temperature, sound speed and ionisation fraction.
        call thermo(tau,cs2,opacity)

!  Compute expansion rate from: grho 8*pi*rho*a**2

        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*a**(-1-3*w_lam)
        grhormass_t=grhormass/a2
        
        if (CP%Num_Nu_Massive == 0) then
           rhonu=1._dl
           pnu=1._dl/3._dl
           clxnu=0._dl
           qnu = 0
        else
           !Get density and pressure as ratio to massles by interpolation from table
           call Nu_background(a*amnu,rhonu,pnu)
           if (CP%MassiveNuMethod == Nu_approx) then

             clxnu=ay(EV%iq0)
             qnu=ay(EV%iq0+1)

           else
           if (EV%MassiveNuApprox) then    
              clxnu=ay(EV%iq0)
              qnu=ay(EV%iq0+2)
           else
            !Integrate over q

            call Nu_Integrate01(a*amnu,clxnu,qnu,ay(EV%iq0),ay(EV%iq1))
            !clxnu_here  = rhonu*clxnu, qnu_here = qnu*rhonu
            qnu=qnu/rhonu
            clxnu = clxnu/rhonu
           endif 
           end if
        end if
        
        grhonu_t=grhormass_t*rhonu
        gpnu_t=grhormass_t*pnu

        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t+grhonu_t
        
        adotoa=sqrt((grho+grhok)/3._dl)
        ayprime(1)=adotoa*a

!  Photon mass density over baryon mass density
        photbar=grhog_t/grhob_t
        pb43=4._dl/3._dl*photbar        
     

!  8*pi*a*a*SUM[rho_i*clx_i]
        dgrho=grhob_t*clxb+grhoc_t*clxc + grhog_t*clxg+grhor_t*clxr +grhonu_t*clxnu

!  8*pi*a*a*SUM[(rho_i+p_i)*v_i]
        dgq=grhob_t*vb+grhog_t*qg+grhor_t*qr+ grhonu_t*qnu

        if (w_lam > -1.and. w_Perturb) then
           clv=ay(EV%w_ix)
           clvdot=ay(EV%w_ix+1) 
           phidot = sqrt((1+w_lam)*grhov_t)
           dgrho=dgrho + (clvdot -clv*1.5_dl*(1-w_lam)*adotoa)*phidot
           dgq=dgq + k*phidot*clv
        end if

!  Get sigma (shear) and z from the constraints
!  have to get z from eta for numerical stability       
        z=(0.5_dl*dgrho/k + etak)/adotoa 
        sigma=(z+1.5_dl*dgq/k2)/EV%Kf(1)
        

  !  ddota/a
        gpres=(grhog_t+grhor_t)/3._dl +gpnu_t +grhov_t*w_lam
        adotdota=0.5_dl*(adotoa*adotoa-gpres)

        if (w_lam > -1.and. w_Perturb ) then
           ayprime(EV%w_ix)= clvdot
! a^2 d^2 V/d phi^2
           Vphiphi = -1.5_dl*(1-w_lam)*(adotdota - (7+3*w_lam)/2*adotoa**2)
           ayprime(EV%w_ix+1) = - 2*adotoa*clvdot - k*z*phidot - k2*clv - clv*Vphiphi
        end if

 
!  eta*k equation
        ayprime(2)=0.5_dl*dgq + curv*z

!  CDM equation of motion
        clxcdot=-k*z
        ayprime(3)=clxcdot
!  Baryon equation of motion.
        clxbdot=-k*(z+vb)
        ayprime(4)=clxbdot
!  Photon equation of motion
        clxgdot=-k*(4._dl/3._dl*z+qg)

!  Drag: a*n_e*sigma_T*(4/3*vb-qg)
        drag=opacity*(4._dl/3._dl*vb-qg)
       
!  use tight coupling approx?
        if (k > 0.06_dl*epsw) then
           ep=ep0
        else
           ep=1.1_dl*ep0
        end if

!     There was an instability for low k when we 
!     switched off tight coupling. Minor change in 
!     criteria ends the problem. (Thanks to David Spergel)
        tcp = ((EV%q/opacity > ep).or.(1._dl/(opacity*tau) > ep .and. EV%q/opacity > 1d-4))          
      
!  Use explicit equation for vb if appropriate
        
        if (tcp) then
           vbdot=-adotoa*vb+cs2*k*clxb-photbar*drag
        else

!  Use tight-coupling approximation for vb
!  zeroth order (in t_c) approximation to vbdot
        vbdot=(-adotoa*vb+cs2*k*clxb  &
             +0.25_dl*k*pb43*(clxg-2._dl*EV%Kf(1)*pig))/(1._dl+pb43) 


!  First-order (in t_c) approximation to baryon-photon splip
        slip=2._dl*pb43/(1._dl+pb43)*adotoa*(vb-3._dl/4._dl*qg) &
           +1._dl/opacity*(-adotdota*vb-0.5_dl*k*adotoa*clxg  &
           +k*(cs2*clxbdot-0.25_dl*clxgdot))/(1._dl+pb43)

!  First-order approximation to vbdot
        vbdot=vbdot+pb43/(1._dl+pb43)*slip
        end if
        ayprime(5)=vbdot

!  Photon equations of motion
        ayprime(6)=clxgdot
        qgdot=4._dl/3._dl*(-vbdot-adotoa*vb+cs2*k*clxb)/pb43 &
            +k*(clxg-2._dl*pig*EV%Kf(1))/3._dl
        ayprime(7)=qgdot


        if (EV%FirstZerolForBeta <= EV%MaxlNeeded) ayprime(8:EV%ScalEqsToPropagate)=0 !bit lazy this

!  Use explicit equations for photon moments if appropriate
        
        polter = 0.1_dl*pig+9._dl/15._dl*ay(EV%polind+2) !2/15*(3/4 pig + 9/2 E2)
        if (tcp) then
        pigdot=0.4_dl*k*qg-0.6_dl*k*EV%Kf(2)*ay(9)-opacity*(pig - polter) &
               +8._dl/15._dl*k*sigma
        ayprime(8)=pigdot
        do  l=3,min(EV%FirstZerolForBeta,EV%lmaxg)-1
           ayprime(l+6)=k*denl(l)*(l*ay(l+5)-(l+1)*EV%Kf(l)*ay(l+7))-opacity*ay(l+6)
        end do
!  Truncate the photon moment expansion
        if (EV%lmaxg/=EV%FirstZerolForBeta) then
          
          ayprime(EV%lmaxg+6)=k*ay(EV%lmaxg+5)-(EV%lmaxg+1)*cothxor*ay(EV%lmaxg+6)  &
                       -opacity*ay(EV%lmaxg+6)
        end if

        else
!  Use tight-coupling approximation where moments are zero for l>1
        pigdot=0._dl
        
        ayprime(8:EV%lmaxg+6)=0._dl
        
        end if


!  Massless neutrino equations of motion.
        clxrdot=-k*(4._dl/3._dl*z+qr)
        ayprime(EV%lmaxg+7)=clxrdot
        qrdot=k*(clxr-2._dl*pir*EV%Kf(1))/3._dl
        ayprime(EV%lmaxg+8)=qrdot
        pirdot=k*(0.4_dl*qr-0.6_dl*EV%Kf(2)*ay(EV%lmaxg+10)+8._dl/15._dl*sigma)
        ayprime(EV%lmaxg+9)=pirdot
        do l=3,min(EV%FirstZerolForBeta,EV%lmaxnr)-1
           ayprime(l+EV%lmaxg+7)=k*denl(l)*(l*ay(l+EV%lmaxg+6) -(l+1)*EV%Kf(l)*ay(l+EV%lmaxg+8))
        end do   
!  Truncate the neutrino expansion  
        if (EV%lmaxnr/=EV%FirstZerolForBeta) then
        
         ayprime(EV%lmaxnr+EV%lmaxg+7)=k*ay(EV%lmaxnr+EV%lmaxg+6)- &
                                (EV%lmaxnr+1)*cothxor*ay(EV%lmaxnr+EV%lmaxg+7)
        end if

!  Polarization
        if (tcp) then     
        !l=2 (defined to be zero for l<2)
         ayprime(EV%polind+2) = -opacity*(ay(EV%polind+2) - polter) - k/3._dl*EV%Kf(2)*ay(EV%polind+3)
        !and the rest
        do l=3,min(EV%FirstZerolForBeta,EV%lmaxgpol)-1
           ayprime(EV%polind+l)=-opacity*ay(EV%polind+l) + k*denl(l)*(l*ay(EV%polind+l-1) -&
                          polfac(l)*EV%Kf(l)*ay(EV%polind+l+1))
        end do  
        
        !truncate
        if (EV%lmaxgpol/=EV%FirstZerolForBeta) then   
         
         ayprime(EV%polind+EV%lmaxgpol)=-opacity*ay(EV%polind+EV%lmaxgpol) +  &
          k*EV%poltruncfac*ay(EV%polind+EV%lmaxgpol-1)-(EV%lmaxgpol+3)*cothxor*ay(EV%polind+EV%lmaxgpol)
        end if
        
        else
         ayprime(EV%polind+2:EV%polind+EV%lmaxgpol)=0._dl
        end if
!  Massive neutrino equations of motion.
        if (CP%Num_Nu_massive == 0) return

         if (CP%MassiveNuMethod == Nu_approx) then

            !crude approximation scheme
            w=pnu/rhonu
            f = (5._dl/3)**(a*amnu/(a*amnu+200))*(3*w)**((2+a*amnu)/(4+a*amnu)) 
         
            !clxnudot
            ayprime(EV%iq0)=-(f-3*w)*adotoa*clxnu-k*((1+w)*z+qnu)
            !qnudot
            ayprime(EV%iq0+1)=-adotoa*(1-3*w)*ay(EV%iq0+1)+k/3._dl*(f*clxnu-2._dl*EV%Kf(1)*ay(EV%iq0+2))
            !pinudot

            ayprime(EV%iq0+2)=-adotoa*(-f +2-3*w)*ay(EV%iq0+2) +  &
                 k*(2._dl/5*f*qnu-0.6_dl*EV%Kf(2)*ay(EV%iq0+3)+ 2._dl/5*w*(5-3*w)*sigma)
 
           do l=3,EV%lmaxnu-1
               ayprime(EV%iq0+l)=-adotoa*((1-l)*f+l-3*w)*ay(EV%iq0+l) + &
                    k*denl(l)*(f*l*ay(EV%iq0+l-1) -EV%Kf(l)*(l+1)*ay(EV%iq0+l+1))
            end do
            !  Truncate the neutrino expansion
            ayprime(EV%iq0+EV%lmaxnu)=k*ay(EV%iq0+EV%lmaxnu-1)- &
                 (EV%lmaxnu+1)*cothxor*ay(EV%iq0+EV%lmaxnu)
                        
         
         else !Not approx scheme

          if (EV%MassiveNuApprox) then
             !Now EV%iq0 = clx, EV%iq0+1 = clxp, EV%iq0+2 = G_1, EV%iq0+3=G_2=pinu
             G11_t=EV%G11/a2/a
             G30_t=EV%G30/a2/a
             w=pnu/rhonu
             ayprime(EV%iq0)=-k*z*(w+1) + 3*adotoa*(w*ay(EV%iq0) - ay(EV%iq0+1))-k*ay(EV%iq0+2)
             ayprime(EV%iq0+1)=(3*w-2)*adotoa*ay(EV%iq0+1) - 5._dl/3*k*z*w - k/3*G11_t
             ayprime(EV%iq0+2)=(3*w-1)*adotoa*ay(EV%iq0+2) - k*(2._dl/3*EV%Kf(1)*ay(EV%iq0+3)-ay(EV%iq0+1))
             ayprime(EV%iq0+3)=(3*w-2)*adotoa*ay(EV%iq0+3) + 2*w*k*sigma - k/5*(3*EV%Kf(2)*G30_t-2*G11_t)
     
          else
          
          
          do i=1,nqmax
            q=i*dq-0.5_dl
            aq=a*amnu/q
            v=1._dl/sqrt(1._dl+aq*aq)
            akv(i)=k*v
          end do
!  l = 0, 1, 2,EV%lmaxnu.
          do i=1,nqmax             
           ind=EV%iq0+i-1
           ayprime(ind)=-akv(i)*ay(ind+nqmax)+z*k*dlfdlq(i)/3
           ind=EV%iq1+i-1
           ayprime(ind)=akv(i)*(ay(ind-nqmax)-2*EV%Kf(1)*ay(ind+nqmax))/3
           ind=EV%iq2+i-1
           ayprime(ind)=akv(i)*(2*ay(ind-nqmax)-3*EV%Kf(2)*ay(ind+nqmax))/5 &
                 -k*2._dl/15._dl*sigma*dlfdlq(i)
           ind=EV%iq0+i-1+EV%lmaxnu*nqmax
!  Truncate moment expansion.
           if (EV%lmaxnu<EV%FirstZerolForBeta) then
             ayprime(ind)=akv(i)*ay(ind-nqmax)-(EV%lmaxnu+1)*cothxor*ay(ind)
        
           end if         

          end do
          do l=3,min(EV%FirstZerolForBeta,EV%lmaxnu)-1
            do i=1,nqmax
            ind=EV%iq0-1+i+l*nqmax
            ayprime(ind)=akv(i)*denl(l)*(l*ay(ind-nqmax)-(l+1)*EV%Kf(l)*ay(ind+nqmax))
            end do
          end do  
             
       
          end if     

          end if
        end subroutine derivs

!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc 
        subroutine derivst(EV,n,tau,ayt,aytprime)
!  Evaluate the time derivatives of the tensor perturbations.
        use ThermoData
        use MassiveNu
     
        implicit none      
        type(EvolutionVars) EV
        integer n,l,i,ind
        real(dl), target ::  ayt(n),aytprime(n)
        real(dl) ep,tau,grho,rhopi,cs2,opacity,gpres,pirdt,cothxor,grhormass_t
        logical tcp
        real(dl), dimension(:),pointer :: neut,neutprime,E,B,Eprime,Bprime
        real(dl) q,aq,v,akv(nqmax0),shearnu
        real(dl)  grhob_t,grhor_t,grhoc_t,grhog_t,grhov_t,grhonu_t,gpnu_t,polter
        real(dl) Hchi,shear,aux, pig
        real(dl) k,k2,a,a2
        real(dl) pinu,pir,pnu,adotoa, rhonu
      
        k2=EV%k2_buf
        k= EV%k_buf   
        aux=EV%aux_buf 


        !E and B start at l=2. Set up pointers accordingly to fill in ayt arrays
        E => ayt(EV%lmaxt+3-1:)
        Eprime=> aytprime(EV%lmaxt+3-1:) 
        B => E(EV%lmaxpolt:)
        Bprime => Eprime(EV%lmaxpolt:)
      
        cothxor=1._dl/tanfunc(tau/r)/r  
       
        a=ayt(1)        

!  Electric part of Weyl tensor and shear tensor
!        elec=ayt(2)

        Hchi=ayt(2)  

        shear=ayt(3)
!  Photon anisotropic stress
        pig=ayt(4)

        a2=a*a

!  Get sound speed and opacity, and see if should use tight-coupling
             
        call thermo(tau,cs2,opacity)
        if (k > 0.06_dl*epsw) then
           ep=ep0
        else
           ep=0.2_dl*ep0
        end if
        tcp = ((k/opacity > ep).or.(1._dl/(opacity*tau) > ep)) 
        
       if (CP%Num_Nu_Massive == 0) then
           rhonu=1._dl
           pnu=1._dl/3._dl
        else
           call Nu_background(a*amnu,rhonu,pnu) 
        end if
! Compute expansion rate from: grho=8*pi*rho*a**2
! Also calculate gpres: 8*pi*p*a**2
        grhob_t=grhob/a
        grhoc_t=grhoc/a
        grhor_t=grhornomass/a2
        grhog_t=grhog/a2
        grhov_t=grhov*a**(-1-3*w_lam)

        grhormass_t=grhormass/a2

        grhonu_t=grhormass_t*rhonu
        gpnu_t=grhormass_t*pnu

        grho=grhob_t+grhoc_t+grhor_t+grhog_t+grhov_t+grhonu_t
        
        gpres=(grhog_t+grhor_t)/3._dl+grhov_t*w_lam +gpnu_t
     
        adotoa=sqrt((grho+grhok)/3._dl)
        aytprime(1)=adotoa*a
        polter = 0.1_dl*pig + 9._dl/15._dl*E(2)

        if (tcp) then
!  Don't use tight coupling approx - use explicit equations:
!  Equation for the photon anisotropic stress
        aytprime(4)=k*(-1._dl/3._dl*EV%Kft(2)*ayt(5)+8._dl/15._dl*shear)  &
                  -opacity*(pig - polter)


        do l=3,min(EV%FirstZerolForBeta,EV%lmaxt)-1
         aytprime(l+2)=k*denl(l)*(l*ayt(l+1)-   &
                  tensfac(l)*EV%Kft(l)*ayt(l+3))-opacity*ayt(l+2)
 
        end do

        !Truncate the hierarchy 
        if (EV%lmaxt/=EV%FirstZerolForBeta) then
         aytprime(EV%lmaxt+2)=k*EV%lmaxt/(EV%lmaxt-2._dl)*ayt(EV%lmaxt+1)- &
                       (EV%lmaxt+3._dl)*cothxor*ayt(EV%lmaxt+2)-opacity*ayt(EV%lmaxt+2)
        end if

!E and B-bar equations
     
        Eprime(2) = - opacity*(E(2) - polter) + k*(4._dl/6._dl*aux*B(2) - &
                        5._dl/27._dl*EV%Kft(2)*E(3))
        
        do l=3,min(EV%FirstZerolForBeta,EV%lmaxpolt)-1
           Eprime(l) =-opacity*E(l) + k*(denl(l)*(l*E(l-1) - &
                        tensfacpol(l)*EV%Kft(l)*E(l+1)) + 4._dl/(l*(l+1))*aux*B(l))
   
       end do

!truncate: difficult, but zetting to zero seems to work OK
        Eprime(EV%lmaxpolt)=0._dl
                    
        do l=2,min(EV%FirstZerolForBeta,EV%lmaxpolt)-1
        Bprime(l) =-opacity*B(l) + k*(denl(l)*(l*B(l-1) - &
                   tensfacpol(l)*EV%Kft(l)*B(l+1)) - 4._dl/(l*(l+1))*aux*E(l))
        end do

!truncate
        Bprime(EV%lmaxpolt)=0._dl
       
        else  !Tight coupling
        ayt(4)=32._dl/45._dl*k/opacity*shear  
        E(2)=ayt(4)/4._dl
        aytprime(4:n)=0._dl
        
        endif
      

!  Neutrino equations: 
!  Anisotropic stress
        if (DoTensorNeutrinos) then
        
        neutprime => Bprime(EV%lmaxpolt:)
        neut => B(EV%lmaxpolt:)
        
       
!  Massless neutrino anisotropic stress
        pir=neut(2)

        pirdt=k*(-1._dl/3._dl*EV%Kft(2)*neut(3)+ 8._dl/15._dl*shear)
        neutprime(2)=pirdt
!  And for the moments
        do  l=3,min(EV%FirstZerolForBeta,EV%lmaxnrt)-1
           neutprime(l)=k*denl(l)*(l*neut(l-1)- tensfac(l)*EV%Kft(l)*neut(l+1))
        end do
        
!  Truncate the hierarchy
        if (EV%lmaxnrt/=EV%FirstZerolForBeta) then
        neutprime(EV%lmaxnrt)=k*EV%lmaxnrt/(EV%lmaxnrt-2._dl)*neut(EV%lmaxnrt-1)-  &
                       (EV%lmaxnrt+3._dl)*cothxor*neut(EV%lmaxnrt)
        endif


         !  Massive neutrino equations of motion.
         if (CP%Num_Nu_massive/= 0) then
          
          do i=1,nqmax
            q=i*dq-0.5_dl
            aq=a*amnu/q
            v=1._dl/sqrt(1._dl+aq*aq)
            akv(i)=k*v
          end do
          do i=1,nqmax
           ind=EV%iqt+i-1                          
           aytprime(ind)=-1._dl/3._dl*akv(i)*EV%Kft(2)*ayt(ind+nqmax)- 2._dl/15._dl*k*shear*dlfdlq(i)
           ind=EV%iqt+i-1+(EV%lmaxnut-2)*nqmax
!  Truncate moment expansion.
           if (EV%lmaxnut/=EV%FirstZerolForBeta) then
           aytprime(ind)=akv(i)*EV%lmaxnut/(EV%lmaxnut-2._dl)*ayt(ind-nqmax)-(EV%lmaxnut+3)*cothxor*ayt(ind)
           end if         
          end do
          do l=3,min(EV%FirstZerolForBeta,EV%lmaxnut)-1
            do i=1,nqmax
            ind=EV%iqt-1+i+(l-2)*nqmax
            aytprime(ind)=akv(i)*denl(l)*(l*ayt(ind-nqmax)-tensfac(l)*EV%Kft(l)*ayt(ind+nqmax))     
           end do
          end do

           call Nu_Shear(a*amnu,shearnu,ayt(EV%iqt))                 
           pinu=1.5_dl*shearnu/rhonu   
       
         else
            pinu=0._dl
       
         end if

                  
        else
          pinu=0
          pir=0
          pirdt=0  
       
        end if
      
!  Get the propagation equation for the shear
        rhopi=grhog_t*pig+grhor_t*pir+grhonu_t*pinu
    !    aytprime(3)=-adotoa*shear-k*elec-0.5_dl/k*rhopi   
        
        aytprime(3)=-2*adotoa*shear+k*Hchi*(1+2*curv/k2)-rhopi/k   


!  And the electric part of the Weyl.
        if (tcp) then
!  Use the full expression for pigdt
           EV%tenspigdot=aytprime(4)
        else
!  Use the tight-coupling approximation
           EV%tenspigdot=32._dl/45._dl*k/opacity*(2._dl*adotoa*shear+aytprime(3))
        endif

        aytprime(2)=-k*shear

        end subroutine derivst



!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

        end module GaugeInterface
